@@ -5,6 +5,7 @@
  * Copyright (c) 2024-2025 Emil Tsalapatis <etsal@meta.com>
  */
 
+#include "scxtest/scx_test.h"
 #include <scx/common.bpf.h>
 #include <lib/sdt_task.h>
 #include <scx/arena_userspace_interrop.bpf.h>
@@ -1210,7 +1211,7 @@ u64 scx_buddy_chunk_alloc(scx_buddy_chunk_t *chunk, int order_req)
 {
 	scx_buddy_header_t *header;
 	u64 address;
-	u64 order;
+	u64 order = 0;
 	u32 idx;
 
 	bpf_for(order, order_req, SCX_BUDDY_CHUNK_MAX_ORDER) {
