@@ -7,6 +7,7 @@ use anyhow::{anyhow, bail, Context, Result};
 use libbpf_rs::libbpf_sys::*;
 use libbpf_rs::{AsRawLibbpf, OpenProgramImpl};
 use log::warn;
+use std::env;
 use std::ffi::c_void;
 use std::ffi::CStr;
 use std::ffi::CString;
@@ -40,6 +41,9 @@ lazy_static::lazy_static! {
         read_enum("scx_pick_idle_cpu_flags", "SCX_PICK_IDLE_CORE").unwrap_or(0);
     pub static ref SCX_PICK_IDLE_IN_NODE: u64 =
         read_enum("scx_pick_idle_cpu_flags", "SCX_PICK_IDLE_IN_NODE").unwrap_or(0);
+
+    pub static ref ROOT_PREFIX: String =
+        env::var("SCX_SYSFS_PREFIX").unwrap_or("".to_string());
 }
 
 fn load_vmlinux_btf() -> &'static mut btf {
@@ -239,6 +243,25 @@ pub fn cond_kprobe_enable<T>(sym: &str, prog_ptr: &OpenProgramImpl<T>) -> Result
     Ok(false)
 }
 
+pub fn cond_kprobes_enable<T>(kprobes: Vec<(&str, &OpenProgramImpl<T>)>) -> Result<bool> {
+    // Check if all the symbols exist.
+    for (sym, _) in kprobes.iter() {
+        if in_kallsyms(sym)? == false {
+            warn!("symbol {} is missing, kprobe not loaded", sym);
+            return Ok(false);
+        }
+    }
+
+    // Enable all the tracepoints.
+    for (_, ptr) in kprobes.iter() {
+        unsafe {
+            bpf_program__set_autoload(ptr.as_libbpf_object().as_ptr(), true);
+        }
+    }
+
+    Ok(true)
+}
+
 pub fn cond_tracepoint_enable<T>(tracepoint: &str, prog_ptr: &OpenProgramImpl<T>) -> Result<bool> {
     if tracepoint_exists(tracepoint)? {
         unsafe {
@@ -254,6 +277,26 @@ pub fn cond_tracepoint_enable<T>(tracepoint: &str, prog_ptr: &OpenProgramImpl<T>
 
     Ok(false)
 }
+
+pub fn cond_tracepoints_enable<T>(tracepoints: Vec<(&str, &OpenProgramImpl<T>)>) -> Result<bool> {
+    // Check if all the tracepoints exist.
+    for (tp, _) in tracepoints.iter() {
+        if tracepoint_exists(tp)? == false {
+            warn!("tradepoint {} is missing, tracepoint not loaded", tp);
+            return Ok(false);
+        }
+    }
+
+    // Enable all the tracepoints.
+    for (_, ptr) in tracepoints.iter() {
+        unsafe {
+            bpf_program__set_autoload(ptr.as_libbpf_object().as_ptr(), true);
+        }
+    }
+
+    Ok(true)
+}
+
 pub fn is_sched_ext_enabled() -> io::Result<bool> {
     let content = std::fs::read_to_string("/sys/kernel/sched_ext/state")?;
 
