@@ -3,8 +3,9 @@
 use crate::misc::read_from_file;
 use crate::{Cpumask, NR_CPU_IDS};
 use nvml_wrapper::bitmasks::InitFlags;
-use nvml_wrapper::enum_wrappers::device::{Clock, TopologyLevel};
+use nvml_wrapper::enum_wrappers::device::{Clock, PerformanceState, TopologyLevel};
 use nvml_wrapper::Nvml;
+use nvml_wrapper_sys::bindings::NVML_AFFINITY_SCOPE_NODE;
 use std::collections::BTreeMap;
 use std::path::Path;
 
@@ -20,12 +21,20 @@ pub struct Gpu {
     pub max_graphics_clock: usize,
     // AMD uses CU for this value
     pub max_sm_clock: usize,
+    // Frequency of the GPU's memory
+    pub max_mem_clock: usize,
+    // Streaming Multiprocessor count
+    pub multiproc_count: usize,
     pub memory: u64,
     pub cpu_mask: Cpumask,
     // Represents the ordered list of nearest
     // available devices in term of topology
     // connectivity (as for now in term of PCI board).
     pub nearest: Vec<GpuIndex>,
+    // Current (P)State which determines the
+    // performance level/energy consumption ratio
+    // starting with Zero being the highest.
+    pub perf_state: PerformanceState,
 }
 
 pub fn create_gpus() -> BTreeMap<usize, Vec<Gpu>> {
@@ -44,6 +53,9 @@ pub fn create_gpus() -> BTreeMap<usize, Vec<Gpu>> {
                 .max_customer_boost_clock(Clock::Graphics)
                 .unwrap_or(0);
             let sm_boost_clock = nvidia_gpu.max_customer_boost_clock(Clock::SM).unwrap_or(0);
+            let mem_boost_clock = nvidia_gpu
+                .max_customer_boost_clock(Clock::Memory)
+                .unwrap_or(0);
             let Ok(memory_info) = nvidia_gpu.memory_info() else {
                 continue;
             };
@@ -54,7 +66,9 @@ pub fn create_gpus() -> BTreeMap<usize, Vec<Gpu>> {
                 continue;
             };
 
-            let cpu_mask = if let Ok(cpu_affinity) = nvidia_gpu.cpu_affinity(*NR_CPU_IDS) {
+            let cpu_mask = if let Ok(cpu_affinity) =
+                nvidia_gpu.cpu_affinity_within_scope(*NR_CPU_IDS, NVML_AFFINITY_SCOPE_NODE)
+            {
                 // Note: nvml returns it as an arch dependent array of integrals
                 #[cfg(target_pointer_width = "32")]
                 let cpu_affinity: Vec<u64> = cpu_affinity
@@ -66,6 +80,12 @@ pub fn create_gpus() -> BTreeMap<usize, Vec<Gpu>> {
                 Cpumask::new()
             };
 
+            let multiproc_count = if let Ok(attributes) = nvidia_gpu.attributes() {
+                attributes.multiprocessor_count
+            } else {
+                0
+            };
+
             let nearest_gpu_topology_level = if nvidia_gpu.is_multi_gpu_board().unwrap_or(false) {
                 // e.g. for some Tesla models, this mode is faster
                 // as units are part of the same physical device
@@ -94,6 +114,10 @@ pub fn create_gpus() -> BTreeMap<usize, Vec<Gpu>> {
                 Vec::new()
             };
 
+            let perf_state = nvidia_gpu
+                .performance_state()
+                .unwrap_or(PerformanceState::Unknown);
+
             // The NVML library doesn't return a PCIe bus ID compatible with sysfs. It includes
             // uppercase bus ID values and an extra four leading 0s.
             let bus_id = pci_info.bus_id.to_lowercase();
@@ -106,17 +130,14 @@ pub fn create_gpus() -> BTreeMap<usize, Vec<Gpu>> {
                 node_id: numa_node as usize,
                 max_graphics_clock: graphics_boost_clock as usize,
                 max_sm_clock: sm_boost_clock as usize,
+                max_mem_clock: mem_boost_clock as usize,
+                multiproc_count: multiproc_count as usize,
                 memory: memory_info.total,
                 cpu_mask,
                 nearest,
+                perf_state,
             };
-            if !gpus.contains_key(&numa_node) {
-                gpus.insert(numa_node, vec![gpu]);
-                continue;
-            }
-            if let Some(gpus) = gpus.get_mut(&numa_node) {
-                gpus.push(gpu);
-            }
+            gpus.entry(gpu.node_id).or_default().push(gpu);
         }
     }
 
