@@ -12,12 +12,15 @@ use stats::Metrics;
 use scx_p2dq::SchedulerOpts as P2dqOpts;
 use scx_userspace_arena::alloc::Allocator;
 use scx_userspace_arena::alloc::HeapAllocator;
+use scx_utils::build_id;
+use scx_utils::compat::tracefs_mount;
 use scx_utils::init_libbpf_logging;
 use scx_utils::scx_ops_attach;
 use scx_utils::scx_ops_load;
 use scx_utils::scx_ops_open;
 use scx_utils::uei_exited;
 use scx_utils::uei_report;
+
 use scx_utils::Core;
 use scx_utils::Llc;
 use scx_utils::Topology;
@@ -32,6 +35,9 @@ use anyhow::Context;
 use anyhow::Result;
 use clap::Parser;
 use crossbeam::channel::RecvTimeoutError;
+use libbpf_rs::libbpf_sys::bpf_program__set_autoattach;
+use libbpf_rs::AsRawLibbpf;
+use libbpf_rs::Link;
 use libbpf_rs::MapCore as _;
 use libbpf_rs::OpenObject;
 use libbpf_rs::ProgramInput;
@@ -41,6 +47,9 @@ use nix::unistd::Pid;
 use scx_stats::prelude::*;
 
 use std::alloc::Layout;
+use std::collections::HashSet;
+use std::fs::File;
+use std::io::{BufRead, BufReader};
 use std::marker::PhantomPinned;
 use std::mem::MaybeUninit;
 use std::panic;
@@ -127,11 +136,18 @@ pub enum RequiresPpid {
     IncludeParent(Pid),
 }
 
+#[derive(Debug)]
+pub struct KprobeRandomDelays {
+    pub kprobes: Vec<String>,
+    pub freq: f64,
+}
+
 #[derive(Debug)]
 /// State required to build a Scheduler configuration.
 pub struct Builder<'a> {
     pub traits: Vec<Trait>,
     pub verbose: u8,
+    pub kprobe_random_delays: Option<KprobeRandomDelays>,
     pub p2dq_opts: &'a P2dqOpts,
     pub requires_ppid: Option<RequiresPpid>,
 }
@@ -147,6 +163,7 @@ pub struct SkelWithObject {
 pub struct Scheduler {
     _arena: HeapAllocator<ArenaAllocator>,
     _struct_ops: libbpf_rs::Link,
+    _links: Vec<Link>,
     stats_server: StatsServer<(), Metrics>,
 
     // Fields are dropped in declaration order, this must be last as arena holds a reference to the
@@ -347,6 +364,29 @@ impl Builder<'_> {
         Ok(())
     }
 
+    fn attach_kprobes(&self, skel: &mut BpfSkel) -> Result<Vec<Link>> {
+        let Some(kd) = &self.kprobe_random_delays else {
+            return Ok(vec![]);
+        };
+
+        if kd.kprobes.is_empty() {
+            return Ok(vec![]);
+        }
+
+        validate_kprobes(&kd.kprobes).context("Failed to validate kprobes passed by user")?;
+
+        let mut links = vec![];
+        for k in &kd.kprobes {
+            links.push(
+                skel.progs
+                    .generic
+                    .attach_kprobe(false, k)
+                    .context(format!("Failed to attach kprobe {:?}", k))?,
+            );
+        }
+        Ok(links)
+    }
+
     fn load_skel(&self) -> Result<Pin<Rc<SkelWithObject>>> {
         let mut out: Rc<MaybeUninit<SkelWithObject>> = Rc::new_uninit();
         let uninit_skel = Rc::get_mut(&mut out).expect("brand new rc should be unique");
@@ -393,6 +433,11 @@ impl Builder<'_> {
             }
         };
 
+        if let Some(kprobe_random_delays) = &self.kprobe_random_delays {
+            open_skel.maps.rodata_data.kprobe_delays_freq_frac32 =
+                (kprobe_random_delays.freq * 2_f64.powf(32_f64)) as u32;
+        }
+
         // Set up the frequency array. The first element means nothing, so should be what's
         // required to add up to 100%. The rest should be cumulative frequencies.
         let freq_array = &mut open_skel.maps.rodata_data.trait_delay_freq_frac32;
@@ -456,6 +501,12 @@ impl Builder<'_> {
             }
         }
 
+        // For now, we'll do it in this way. However, once we upgrade to libbpf_rs 0.25.0,
+        // we can use the set_autoattach method on the OpenProgramImpl to do this.
+        unsafe {
+            bpf_program__set_autoattach(open_skel.progs.generic.as_libbpf_object().as_ptr(), false)
+        };
+
         let mut skel = scx_ops_load!(open_skel, chaos, uei)?;
         scx_p2dq::init_skel!(&mut skel);
 
@@ -485,16 +536,18 @@ impl<'a> TryFrom<Builder<'a>> for Scheduler {
 
         let arena = HeapAllocator::new(ArenaAllocator(skel.clone()));
         let stats_server = StatsServer::new(stats::server_data()).launch()?;
-
-        let struct_ops = {
+        let (links, struct_ops) = {
             let mut skel_guard = skel.skel.write().unwrap();
-            scx_ops_attach!(skel_guard, chaos)?
+            let struct_ops = scx_ops_attach!(skel_guard, chaos)?;
+            let links = b.attach_kprobes(&mut skel_guard)?;
+            (links, struct_ops)
         };
         debug!("scx_chaos scheduler started");
 
         Ok(Scheduler {
             _arena: arena,
             _struct_ops: struct_ops,
+            _links: links,
             stats_server,
             skel,
         })
@@ -528,7 +581,7 @@ pub struct CpuFreqArgs {
     #[clap(long, requires = "cpufreq_frequency")]
     pub cpufreq_min: Option<u32>,
 
-    /// Minimum CPU frequency for scaling.
+    /// Maximum CPU frequency for scaling.
     #[clap(long, requires = "cpufreq_min")]
     pub cpufreq_max: Option<u32>,
 }
@@ -545,6 +598,18 @@ pub struct PerfDegradationArgs {
     pub degradation_frac7: u64,
 }
 
+/// Delay a process when a kprobe is hit.
+#[derive(Debug, Parser)]
+pub struct KprobeArgs {
+    /// Introduce random delays in the scheduler whenever a provided kprobe is hit.
+    #[clap(long, num_args = 1.., value_parser)]
+    pub kprobes_for_random_delays: Vec<String>,
+
+    /// Chance of kprobe random delays. Must be between 0 and 1.
+    #[clap(long, requires = "kprobes_for_random_delays")]
+    pub kprobe_random_delay_frequency: Option<f64>,
+}
+
 /// scx_chaos: A general purpose sched_ext scheduler designed to amplify race conditions
 ///
 /// WARNING: This scheduler is a very early alpha, and hasn't been production tested yet. The CLI
@@ -580,6 +645,10 @@ pub struct Args {
     #[clap(short = 'v', long, action = clap::ArgAction::Count)]
     pub verbose: u8,
 
+    /// Print version and exit.
+    #[clap(long)]
+    pub version: bool,
+
     /// Enable stats monitoring with the specified interval.
     #[clap(long)]
     pub stats: Option<f64>,
@@ -598,6 +667,9 @@ pub struct Args {
     #[command(flatten, next_help_heading = "CPU Frequency")]
     pub cpu_freq: CpuFreqArgs,
 
+    #[command(flatten, next_help_heading = "Kprobe Random Delays")]
+    pub kprobe_random_delays: KprobeArgs,
+
     #[command(flatten, next_help_heading = "General Scheduling")]
     pub p2dq: P2dqOpts,
 
@@ -681,9 +753,21 @@ impl<'a> Iterator for BuilderIterator<'a> {
                 None
             };
 
+            let kprobe_random_delays = match &self.args.kprobe_random_delays {
+                KprobeArgs {
+                    kprobes_for_random_delays,
+                    kprobe_random_delay_frequency,
+                } if !kprobes_for_random_delays.is_empty() => Some(KprobeRandomDelays {
+                    kprobes: kprobes_for_random_delays.clone(),
+                    freq: kprobe_random_delay_frequency.unwrap_or(0.1),
+                }),
+                _ => None,
+            };
+
             Some(Builder {
                 traits,
                 verbose: self.args.verbose,
+                kprobe_random_delays,
                 p2dq_opts: &self.args.p2dq,
                 requires_ppid,
             })
@@ -691,7 +775,40 @@ impl<'a> Iterator for BuilderIterator<'a> {
     }
 }
 
+pub fn validate_kprobes(kprobes: &[String]) -> Result<()> {
+    let path = tracefs_mount()?;
+    let file = File::open(path.join("available_filter_functions"))?;
+    let reader = BufReader::new(file);
+
+    let available_kprobes: HashSet<_> = reader
+        .lines()
+        .filter_map(|line| line.ok()?.split_whitespace().next().map(String::from))
+        .collect();
+
+    let missing: Vec<_> = kprobes
+        .iter()
+        .filter(|probe| !available_kprobes.contains(*probe))
+        .collect();
+
+    if !missing.is_empty() {
+        return Err(anyhow::anyhow!(
+            "The following kprobes are not available: {:?}",
+            missing
+        ));
+    }
+
+    Ok(())
+}
+
 pub fn run(args: Args) -> Result<()> {
+    if args.version {
+        println!(
+            "scx_chaos: {}",
+            build_id::full_version(env!("CARGO_PKG_VERSION"))
+        );
+        return Ok(());
+    }
+
     let args = Arc::new(args);
 
     let shutdown = Arc::new((Mutex::new(false), Condvar::new()));
@@ -706,6 +823,11 @@ pub fn run(args: Args) -> Result<()> {
     })
     .context("Error setting Ctrl-C handler")?;
 
+    info!(
+        "Running scx_chaos (build ID: {})",
+        build_id::full_version(env!("CARGO_PKG_VERSION"))
+    );
+
     if let Some(intv) = args.monitor {
         return stats::monitor(Duration::from_secs_f64(intv), shutdown);
     }
@@ -798,14 +920,16 @@ pub fn run(args: Args) -> Result<()> {
         cvar.notify_all();
     }
 
-    if let Err(e) = scheduler_thread.join() {
-        panic::resume_unwind(e);
+    match scheduler_thread.join() {
+        Ok(Ok(())) => {}
+        Ok(Err(e)) => return Err(e),
+        Err(e) => panic::resume_unwind(e),
     }
 
-    if let Some(stats_thread) = stats_thread {
-        if let Err(e) = stats_thread.join() {
-            panic::resume_unwind(e);
-        }
+    match stats_thread.map(|t| t.join()).unwrap_or(Ok(Ok(()))) {
+        Ok(Ok(())) => {}
+        Ok(Err(e)) => return Err(e),
+        Err(e) => panic::resume_unwind(e),
     }
 
     Ok(())
