@@ -194,7 +194,7 @@ char _license[] SEC("license") = "GPL";
 /*
  * Logical current clock
  */
-static u64		cur_logical_clk;
+static u64		cur_logical_clk = LAVD_DL_COMPETE_WINDOW;
 
 /*
  * Current service time
@@ -219,216 +219,7 @@ const volatile u64	slice_max_ns = LAVD_SLICE_MAX_NS_DFL;
 #include "idle.bpf.c"
 #include "balance.bpf.c"
 #include "sys_stat.bpf.c"
-
-static u32 calc_greedy_ratio(struct task_ctx *taskc)
-{
-	u32 ratio;
-
-	/*
-	 * The greedy ratio of a task represents how much time the task
-	 * overspent CPU time compared to the ideal, fair CPU allocation. It is
-	 * the ratio of task's actual service time to average service time in a
-	 * system.
-	 */
-	ratio = (taskc->svc_time << LAVD_SHIFT) / sys_stat.avg_svc_time;
-	taskc->is_greedy = ratio > LAVD_SCALE;
-	return ratio;
-}
-
-static u32 calc_greedy_factor(u32 greedy_ratio)
-{
-	/*
-	 * For all under-utilized tasks, we treat them equally.
-	 */
-	if (greedy_ratio <= LAVD_SCALE)
-		return LAVD_SCALE;
-
-	/*
-	 * For over-utilized tasks, we give some mild penalty.
-	 */
-	return LAVD_SCALE + ((greedy_ratio - LAVD_SCALE) / LAVD_LC_GREEDY_PENALTY);
-}
-
-static inline u64 calc_runtime_factor(u64 runtime)
-{
-	return rsigmoid_u64(runtime, LAVD_LC_RUNTIME_MAX);
-}
-
-static inline u64 calc_freq_factor(u64 freq)
-{
-	return sigmoid_u64(freq, LAVD_LC_FREQ_MAX);
-}
-
-static u64 calc_weight_factor(struct task_struct *p, struct task_ctx *taskc)
-{
-	u64 weight_boost = 1;
-	u64 weight_ft;
-
-	/*
-	 * Prioritize a wake-up task since this is a clear sign of immediate
-	 * consumer. If it is a synchronous wakeup, double the prioritization.
-	 */
-	weight_boost += taskc->wakeup_ft * LAVD_LC_WEIGHT_BOOST;
-
-	/*
-	 * Prioritize a kernel task since many kernel tasks serve
-	 * latency-critical jobs.
-	 */
-	if (is_kernel_task(p))
-		weight_boost += LAVD_LC_WEIGHT_BOOST;
-
-	/*
-	 * Further prioritize kworkers.
-	 */
-	if (is_kernel_worker(p))
-		weight_boost += LAVD_LC_WEIGHT_BOOST;
-
-	/*
-	 * Prioritize an affinitized task since it has restrictions
-	 * in placement so it tends to be delayed.
-	 */
-	if (taskc->is_affinitized)
-		weight_boost += LAVD_LC_WEIGHT_BOOST;
-
-	/*
-	 * Prioritize a pinned task since it has restrictions in placement
-	 * so it tends to be delayed.
-	 */
-	if (is_pinned(p) || is_migration_disabled(p))
-		weight_boost += LAVD_LC_WEIGHT_BOOST;
-
-	/*
-	 * Prioritize a lock holder for faster system-wide forward progress.
-	 */
-	if (taskc->need_lock_boost) {
-		taskc->need_lock_boost = false;
-		weight_boost += LAVD_LC_WEIGHT_BOOST;
-	}
-
-	/*
-	 * Respect nice priority.
-	 */
-	weight_ft = p->scx.weight * weight_boost;
-	return weight_ft;
-}
-
-static void calc_perf_cri(struct task_struct *p, struct task_ctx *taskc)
-{
-	u64 wait_freq_ft, wake_freq_ft, perf_cri = LAVD_SCALE;
-
-	/*
-	 * A task is more CPU-performance sensitive when it meets the following
-	 * conditions:
-	 *
-	 * - It is in the middle of the task graph (high wait and wake
-	 *   frequencies).
-	 * - Its runtime and frequency are high;
-	 * - Its nice priority is high;
-	 *
-	 * We use the log-ed value since the raw value follows the highly
-	 * skewed distribution.
-	 *
-	 * Note that we use unadjusted weight to reflect the pure task priority.
-	 */
-	if (have_little_core) {
-		wait_freq_ft = calc_freq_factor(taskc->wait_freq);
-		wake_freq_ft = calc_freq_factor(taskc->wake_freq);
-		perf_cri = log2_u64(wait_freq_ft * wake_freq_ft);
-		perf_cri += log2_u64(max(taskc->run_freq, 1) *
-				     max(taskc->avg_runtime, 1) * p->scx.weight);
-	}
-
-	taskc->perf_cri = perf_cri;
-}
-
-static void calc_lat_cri(struct task_struct *p, struct task_ctx *taskc)
-{
-	u64 weight_ft, wait_freq_ft, wake_freq_ft, runtime_ft;
-	u64 lat_cri;
-
-	/*
-	 * Adjust task's weight based on the scheduling context, such as
-	 * if it is a kernel task, lock holder, etc.
-	 */
-	weight_ft = calc_weight_factor(p, taskc);
-
-	/*
-	 * A task is more latency-critical as its wait or wake frequencies
-	 * (i.e., wait_freq and wake_freq) are higher.
-	 *
-	 * Since those frequencies are unbounded and their upper limits are
-	 * unknown, we transform them using sigmoid-like functions. For wait
-	 * and wake frequencies, we use a sigmoid function (sigmoid_u64), which
-	 * is monotonically increasing since higher frequencies mean more
-	 * latency-critical.
-	 */
-	wait_freq_ft = calc_freq_factor(taskc->wait_freq) + 1;
-	wake_freq_ft = calc_freq_factor(taskc->wake_freq) + 1;
-	runtime_ft = calc_runtime_factor(taskc->avg_runtime) + 1;
-
-	/*
-	 * Wake frequency and wait frequency represent how much a task is used
-	 * for a producer and a consumer, respectively. If both are high, the
-	 * task is in the middle of a task chain. The ratio tends to follow an
-	 * exponentially skewed distribution, so we linearize it using log2. We
-	 * add +1 to guarantee the latency criticality (log2-ed) is always
-	 * positive.
-	 */
-	lat_cri = log2_u64(wait_freq_ft * wake_freq_ft) +
-		  log2_u64(runtime_ft * weight_ft);
-
-	/*
-	 * Determine latency criticality of a task in a context-aware manner by
-	 * considering which task wakes up this task. If its waker is more
-	 * latency-critcial, inherit waker's latency criticality.
-	 */
-	taskc->lat_cri = max(lat_cri, taskc->lat_cri_waker);
-}
-
-static u64 calc_adjusted_runtime(struct task_ctx *taskc)
-{
-	u64 runtime;
-
-	/*
-	 * Prefer a short-running (avg_runtime) and recently woken-up
-	 * (acc_runtime) task. To avoid the starvation of CPU-bound tasks,
-	 * which rarely sleep, limit the impact of acc_runtime.
-	 */
-	runtime = LAVD_ACC_RUNTIME_MAX +
-		  min(taskc->acc_runtime, LAVD_ACC_RUNTIME_MAX);
-
-	return runtime;
-}
-
-static u64 calc_virtual_deadline_delta(struct task_struct *p,
-				       struct task_ctx *taskc)
-{
-	u64 deadline, adjusted_runtime;
-	u32 greedy_ratio, greedy_ft;
-
-	/*
-	 * Calculate the deadline based on runtime,
-	 * latency criticality, and greedy ratio.
-	 */
-	calc_perf_cri(p, taskc);
-	calc_lat_cri(p, taskc);
-	greedy_ratio = calc_greedy_ratio(taskc);
-	greedy_ft = calc_greedy_factor(greedy_ratio);
-	adjusted_runtime = calc_adjusted_runtime(taskc);
-
-	deadline = (adjusted_runtime * greedy_ft) / taskc->lat_cri;
-
-	return deadline;
-}
-
-static u64 calc_time_slice(struct task_ctx *taskc)
-{
-	if (!taskc)
-		return LAVD_SLICE_MAX_NS_DFL;
-
-	taskc->slice_ns = sys_stat.slice;
-	return taskc->slice_ns;
-}
+#include "lat_cri.bpf.c"
 
 static void reset_suspended_duration(struct cpu_ctx *cpuc)
 {
@@ -492,6 +283,15 @@ static void advance_cur_logical_clk(struct task_struct *p)
 	}
 }
 
+static u64 calc_time_slice(struct task_ctx *taskc)
+{
+	if (!taskc)
+		return LAVD_SLICE_MAX_NS_DFL;
+
+	taskc->slice_ns = sys_stat.slice;
+	return taskc->slice_ns;
+}
+
 static void update_stat_for_running(struct task_struct *p,
 				    struct task_ctx *taskc,
 				    struct cpu_ctx *cpuc, u64 now)
@@ -636,18 +436,6 @@ static void update_stat_for_stopping(struct task_struct *p,
 	taskc->lock_holder_xted = false;
 }
 
-static u64 calc_when_to_run(struct task_struct *p, struct task_ctx *taskc)
-{
-	u64 deadline_delta;
-
-	/*
-	 * Before enqueueing a task to a run queue, we should decide when a
-	 * task should be scheduled.
-	 */
-	deadline_delta = calc_virtual_deadline_delta(p, taskc);
-	return READ_ONCE(cur_logical_clk) + deadline_delta;
-}
-
 s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 		   u64 wake_flags)
 {
@@ -720,9 +508,15 @@ void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 
 	/*
 	 * Calculate when a task can be scheduled for how long.
+	 *
+	 * If the task is re-enqueued due to a higher-priority scheduling class
+	 * taking the CPU, we don't need to recalculate the task's deadline and
+	 * timeslice, as the task hasn't yet run.
 	 */
-	taskc->wakeup_ft += !!(enq_flags & SCX_ENQ_WAKEUP);
-	p->scx.dsq_vtime = calc_when_to_run(p, taskc);
+	if (!(enq_flags & SCX_ENQ_REENQ)) {
+		taskc->wakeup_ft += !!(enq_flags & SCX_ENQ_WAKEUP);
+		p->scx.dsq_vtime = calc_when_to_run(p, taskc);
+	}
 	p->scx.slice = calc_time_slice(taskc);
 
 	/*
@@ -986,15 +780,29 @@ void BPF_STRUCT_OPS(lavd_runnable, struct task_struct *p, u64 enq_flags)
 	/*
 	 * When a task @p is wakened up, the wake frequency of its waker task
 	 * is updated. The @current task is a waker and @p is a waiter, which
-	 * is being wakened up now.
+	 * is being wakened up now. This is true only when
+	 * SCX_OPS_ALLOW_QUEUED_WAKEUP is not set. The wake-up operations are
+	 * batch processed with SCX_OPS_ALLOW_QUEUED_WAKEUP, so @current task
+	 * is no longer a waker task.
 	 */
 	if (!(enq_flags & SCX_ENQ_WAKEUP))
 		return;
 
 	/*
-	 * Filter out unrelated tasks.
+	 * Filter out unrelated tasks. We keep track of tasks under the same
+	 * parent process to confine the waker-wakee relationship within
+	 * closely related tasks.
 	 */
+	if (enq_flags & (SCX_ENQ_PREEMPT | SCX_ENQ_REENQ | SCX_ENQ_LAST))
+		return;
+
 	waker = bpf_get_current_task_btf();
+	if ((p->real_parent != waker->real_parent))
+		return;
+
+	if (is_kernel_task(p) != is_kernel_task(waker))
+		return;
+
 	waker_taskc = get_task_ctx(waker);
 	if (!waker_taskc) {
 		/*
@@ -1173,6 +981,8 @@ void BPF_STRUCT_OPS(lavd_cpu_online, s32 cpu)
 	cpu_ctx_init_online(cpuc, cpu, now);
 
 	__sync_fetch_and_add(&nr_cpus_onln, 1);
+	__sync_fetch_and_add(&total_capacity, cpuc->capacity);
+	update_autopilot_high_cap();
 	update_sys_stat();
 }
 
@@ -1194,6 +1004,8 @@ void BPF_STRUCT_OPS(lavd_cpu_offline, s32 cpu)
 	cpu_ctx_init_offline(cpuc, cpu, now);
 
 	__sync_fetch_and_sub(&nr_cpus_onln, 1);
+	__sync_fetch_and_sub(&total_capacity, cpuc->capacity);
+	update_autopilot_high_cap();
 	update_sys_stat();
 }
 
@@ -1523,6 +1335,7 @@ static s32 init_per_cpu_ctx(u64 now)
 	/*
 	 * Initilize CPU info
 	 */
+	one_little_capacity = LAVD_SCALE;
 	bpf_for(cpu, 0, nr_cpu_ids) {
 		if (cpu >= LAVD_CPU_ID_MAX)
 			break;
@@ -1593,8 +1406,12 @@ static s32 init_per_cpu_ctx(u64 now)
 			bpf_cpumask_set_cpu(cpu, turbo);
 			have_turbo_core = true;
 		}
+
+		if (cpuc->capacity < one_little_capacity)
+			one_little_capacity = cpuc->capacity;
 	}
 	default_big_core_scale = (big_capacity << LAVD_SHIFT) / sum_capacity;
+	total_capacity = sum_capacity;
 
 	/*
 	 * Initialize compute domain id.
@@ -1698,9 +1515,9 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(lavd_init)
 		return err;
 
 	/*
-	 * Initialize the low cpu watermark for autopilot mode.
+	 * Initialize the low & high cpu capacity watermarks for autopilot mode.
 	 */
-	init_autopilot_low_util();
+	init_autopilot_caps();
 
 	/*
 	 * Initilize the current logical clock and service time.
