@@ -34,6 +34,9 @@ enum consts {
 enum cell_stat_idx {
 	CSTAT_LOCAL,
 	CSTAT_GLOBAL,
+	CSTAT_LO_FALLBACK_Q,
+	CSTAT_HI_FALLBACK_Q,
+	CSTAT_DEFAULT_Q,
 	CSTAT_AFFN_VIOL,
 	NR_CSTATS,
 };
@@ -41,7 +44,6 @@ enum cell_stat_idx {
 struct cpu_ctx {
 	u64 cstats[MAX_CELLS][NR_CSTATS];
 	u64 cell_cycles[MAX_CELLS];
-	u32 prev_cell;
 	u32 cell;
 };
 
