@@ -33,8 +33,14 @@ enum consts {
 	MSEC_PER_SEC		= 1000ULL,
 	NSEC_PER_SEC		= NSEC_PER_MSEC * MSEC_PER_SEC,
 
+	MIN_SLICE_USEC		= 10ULL,
+	MIN_SLICE_NSEC		= (10ULL * NSEC_PER_USEC),
+
 	LOAD_BALANCE_SLACK	= 20ULL,
 
+	P2DQ_MIG_DSQ		= 1LLU << 60,
+	P2DQ_INTR_DSQ		= 1LLU << 32,
+
 	// kernel definitions
 	CLOCK_BOOTTIME		= 7,
 
@@ -56,6 +62,9 @@ enum stat_idx {
 	P2DQ_STAT_LLC_MIGRATION,
 	P2DQ_STAT_NODE_MIGRATION,
 	P2DQ_STAT_KEEP,
+	P2DQ_STAT_ENQ_LLC,
+	P2DQ_STAT_ENQ_INTR,
+	P2DQ_STAT_ENQ_MIG,
 	P2DQ_STAT_SELECT_PICK2,
 	P2DQ_STAT_DISPATCH_PICK2,
 	P2DQ_STAT_WAKE_PREV,
