@@ -12,15 +12,16 @@ struct cpu_ctx {
 	u32				llc_id;
 	u64				affn_dsq;
 	u32				dsq_index;
-	u64				dsq_id;
 	u64				slice_ns;
 	u32				perf;
 	bool				interactive;
 	bool				is_big;
+	bool				nice_task;
 	u64				ran_for;
 	u32				node_id;
-	u64				affn_max_vtime;
-	u64				dsqs[MAX_DSQS_PER_LLC];
+	u64				intr_dsq;
+	u64				mig_dsq;
+	u64				llc_dsq;
 	u64				max_load_dsq;
 };
 
@@ -31,14 +32,15 @@ struct llc_ctx {
 	u64				vtime;
 	u32				lb_llc_id;
 	u64				last_period_ns;
-	u64				load;
+	u64				dsq;
+	u64				intr_dsq;
+	u64				mig_dsq;
 	u32				index;
-	bool				all_big;
+	u64				load;
 	u64				affn_load;
-	u64				affn_max_vtime;
-	u64				dsqs[MAX_DSQS_PER_LLC];
-	u64				dsq_max_vtime[MAX_DSQS_PER_LLC];
+	u64				intr_load;
 	u64				dsq_load[MAX_DSQS_PER_LLC];
+	bool				all_big;
 	struct bpf_cpumask __kptr	*cpumask;
 	struct bpf_cpumask __kptr	*big_cpumask;
 	struct bpf_cpumask __kptr	*little_cpumask;
@@ -64,6 +66,8 @@ struct task_p2dq {
 	u64 			last_run_at;
 	u64			llc_runs; /* how many runs on the current LLC */
 	int			last_dsq_index;
+	bool			interactive;
+	bool			was_nice;
 
 	/* The task is a workqueue worker thread */
 	bool			is_kworker;
