@@ -2,10 +2,6 @@
 
 // This software may be used and distributed according to the terms of the
 // GNU General Public License version 2.
-mod bpf_skel;
-pub use bpf_skel::*;
-
-pub mod bpf_intf;
 pub mod stats;
 use stats::Metrics;
 
@@ -46,6 +42,9 @@ use bpf_intf::stat_idx_P2DQ_STAT_DIRECT;
 use bpf_intf::stat_idx_P2DQ_STAT_DISPATCH_PICK2;
 use bpf_intf::stat_idx_P2DQ_STAT_DSQ_CHANGE;
 use bpf_intf::stat_idx_P2DQ_STAT_DSQ_SAME;
+use bpf_intf::stat_idx_P2DQ_STAT_ENQ_INTR;
+use bpf_intf::stat_idx_P2DQ_STAT_ENQ_LLC;
+use bpf_intf::stat_idx_P2DQ_STAT_ENQ_MIG;
 use bpf_intf::stat_idx_P2DQ_STAT_IDLE;
 use bpf_intf::stat_idx_P2DQ_STAT_KEEP;
 use bpf_intf::stat_idx_P2DQ_STAT_LLC_MIGRATION;
@@ -54,6 +53,8 @@ use bpf_intf::stat_idx_P2DQ_STAT_SELECT_PICK2;
 use bpf_intf::stat_idx_P2DQ_STAT_WAKE_LLC;
 use bpf_intf::stat_idx_P2DQ_STAT_WAKE_MIG;
 use bpf_intf::stat_idx_P2DQ_STAT_WAKE_PREV;
+use scx_p2dq::bpf_intf;
+use scx_p2dq::bpf_skel::*;
 use scx_p2dq::SchedulerOpts;
 use scx_p2dq::TOPO;
 
@@ -151,6 +152,9 @@ impl<'a> Scheduler<'a> {
             dsq_change: stats[stat_idx_P2DQ_STAT_DSQ_CHANGE as usize],
             same_dsq: stats[stat_idx_P2DQ_STAT_DSQ_SAME as usize],
             keep: stats[stat_idx_P2DQ_STAT_KEEP as usize],
+            enq_llc: stats[stat_idx_P2DQ_STAT_ENQ_LLC as usize],
+            enq_intr: stats[stat_idx_P2DQ_STAT_ENQ_INTR as usize],
+            enq_mig: stats[stat_idx_P2DQ_STAT_ENQ_MIG as usize],
             select_pick2: stats[stat_idx_P2DQ_STAT_SELECT_PICK2 as usize],
             dispatch_pick2: stats[stat_idx_P2DQ_STAT_DISPATCH_PICK2 as usize],
             llc_migrations: stats[stat_idx_P2DQ_STAT_LLC_MIGRATION as usize],
