@@ -12,9 +12,6 @@ use bpf::*;
 
 mod stats;
 use std::collections::BTreeSet;
-use std::collections::HashMap;
-use std::fs::File;
-use std::io::Read;
 use std::io::{self};
 use std::mem::MaybeUninit;
 use std::time::Duration;
@@ -25,14 +22,14 @@ use clap::Parser;
 use libbpf_rs::OpenObject;
 use log::info;
 use log::warn;
+use procfs::process::Process;
 use scx_stats::prelude::*;
+use scx_utils::build_id;
 use scx_utils::UserExitInfo;
 use stats::Metrics;
 
 const SCHEDULER_NAME: &'static str = "RustLand";
 
-const VERSION: &'static str = env!("CARGO_PKG_VERSION");
-
 /// scx_rustland: user-space scheduler written in Rust
 ///
 /// scx_rustland is designed to prioritize interactive workloads over background CPU-intensive
@@ -60,8 +57,7 @@ const VERSION: &'static str = env!("CARGO_PKG_VERSION");
 ///
 /// All the tasks are stored in a BTreeSet (TaskTree), using the deadline as the ordering key.
 /// Once the order of execution is determined all tasks are sent back to the BPF counterpart
-/// (scx_rustland_core) to be dispatched. To keep track of the accumulated execution time and
-/// vruntime, the scheduler maintains a HashMap (TaskInfoMap), indexed by pid.
+/// (scx_rustland_core) to be dispatched.
 ///
 /// The BPF dispatcher is completely agnostic of the particular scheduling policy implemented in
 /// user-space. For this reason developers that are willing to use this scheduler to experiment
@@ -82,6 +78,11 @@ struct Opts {
     #[clap(short = 'S', long, default_value = "1000")]
     slice_us_min: u64,
 
+    /// If set, per-CPU tasks are dispatched directly to their only eligible CPU.
+    /// This can help enforce affinity-based isolation for better performance.
+    #[clap(short = 'l', long, action = clap::ArgAction::SetTrue)]
+    percpu_local: bool,
+
     /// If specified, only tasks which have their scheduling policy set to SCHED_EXT using
     /// sched_setscheduler(2) are switched. Otherwise, all tasks are switched.
     #[clap(short = 'p', long, action = clap::ArgAction::SetTrue)]
@@ -117,35 +118,6 @@ struct Opts {
 // Time constants.
 const NSEC_PER_USEC: u64 = 1_000;
 
-// Basic item stored in the task information map.
-#[derive(Debug)]
-struct TaskInfo {
-    vruntime: u64, // total vruntime of the task
-}
-
-// Task information map: store total execution time and vruntime of each task in the system.
-//
-// TaskInfo objects are stored in the HashMap and they are indexed by pid.
-//
-// Entries are removed when the corresponding task exits.
-//
-// This information is fetched from the BPF section (through the .exit_task() callback) and
-// received by the user-space scheduler via self.bpf.dequeue_task(): a task with a negative .cpu
-// value represents an exiting task, so in this case we can free the corresponding entry in
-// TaskInfoMap (see also Scheduler::drain_queued_tasks()).
-struct TaskInfoMap {
-    tasks: HashMap<i32, TaskInfo>,
-}
-
-// TaskInfoMap implementation: provide methods to get items and update items by pid.
-impl TaskInfoMap {
-    fn new() -> Self {
-        TaskInfoMap {
-            tasks: HashMap::new(),
-        }
-    }
-}
-
 #[derive(Debug, PartialEq, Eq, PartialOrd, Clone)]
 struct Task {
     qtask: QueuedTask, // queued task
@@ -165,52 +137,12 @@ impl Ord for Task {
     }
 }
 
-// Task pool where all the tasks that needs to run are stored before dispatching (ordered by their
-// shortest deadline using a BTreeSet).
-struct TaskTree {
-    tasks: BTreeSet<Task>,
-    task_map: HashMap<i32, Task>, // Map from pid to task
-}
-
-// Task pool methods (push / pop).
-impl TaskTree {
-    fn new() -> Self {
-        TaskTree {
-            tasks: BTreeSet::new(),
-            task_map: HashMap::new(),
-        }
-    }
-
-    // Add an item to the pool (item will be placed in the tree depending on its deadline, items
-    // with the same deadline will be sorted by pid).
-    fn push(&mut self, task: Task) {
-        // Check if task already exists.
-        if let Some(prev_task) = self.task_map.get(&task.qtask.pid) {
-            self.tasks.remove(prev_task);
-        }
-
-        // Insert/update task.
-        self.tasks.insert(task.clone());
-        self.task_map.insert(task.qtask.pid, task);
-    }
-
-    // Pop the first item from the BTreeSet (item with the shortest deadline).
-    fn pop(&mut self) -> Option<Task> {
-        if let Some(task) = self.tasks.pop_first() {
-            self.task_map.remove(&task.qtask.pid);
-            Some(task)
-        } else {
-            None
-        }
-    }
-}
-
 // Main scheduler object
 struct Scheduler<'a> {
     bpf: BpfScheduler<'a>,                  // BPF connector
+    opts: &'a Opts,                         // scheduler options
     stats_server: StatsServer<(), Metrics>, // statistics
-    task_pool: TaskTree,                    // tasks ordered by deadline
-    task_map: TaskInfoMap,                  // map pids to the corresponding task information
+    tasks: BTreeSet<Task>,                  // tasks ordered by deadline
     min_vruntime: u64,                      // Keep track of the minimum vruntime across all tasks
     init_page_faults: u64,                  // Initial page faults counter
     slice_ns: u64,                          // Default time slice (in ns)
@@ -218,7 +150,7 @@ struct Scheduler<'a> {
 }
 
 impl<'a> Scheduler<'a> {
-    fn init(opts: &Opts, open_object: &'a mut MaybeUninit<OpenObject>) -> Result<Self> {
+    fn init(opts: &'a Opts, open_object: &'a mut MaybeUninit<OpenObject>) -> Result<Self> {
         let stats_server = StatsServer::new(stats::server_data()).launch()?;
 
         // Low-level BPF connector.
@@ -230,14 +162,19 @@ impl<'a> Scheduler<'a> {
             true, // Enable built-in idle CPU selection policy
         )?;
 
-        info!("{} scheduler attached", SCHEDULER_NAME);
+        info!(
+            "{} version {} - scx_rustland_core {}",
+            SCHEDULER_NAME,
+            build_id::full_version(env!("CARGO_PKG_VERSION")),
+            scx_rustland_core::VERSION
+        );
 
         // Return scheduler object.
         Ok(Self {
             bpf,
+            opts,
             stats_server,
-            task_pool: TaskTree::new(),
-            task_map: TaskInfoMap::new(),
+            tasks: BTreeSet::new(),
             min_vruntime: 0,
             init_page_faults: 0,
             slice_ns: opts.slice_us * NSEC_PER_USEC,
@@ -278,46 +215,98 @@ impl<'a> Scheduler<'a> {
         ts.as_nanos() as u64
     }
 
+    // Return the total amount of tasks waiting in the user-space scheduler.
+    fn nr_tasks_scheduled(&mut self) -> u64 {
+        self.tasks.len() as u64
+    }
+
+    // Return the total amount of tasks waiting to be consumed by the user-space scheduler.
+    fn nr_tasks_queued(&mut self) -> u64 {
+        *self.bpf.nr_queued_mut()
+    }
+
+    // Return the total amount of tasks that are waiting to be scheduled.
+    fn nr_tasks_waiting(&mut self) -> u64 {
+        self.nr_tasks_queued() + self.nr_tasks_scheduled()
+    }
+
+    // Return a value inversely proportional to the task's weight.
+    fn scale_by_task_weight_inverse(task: &QueuedTask, value: u64) -> u64 {
+        value * 100 / task.weight
+    }
+
     // Update task's vruntime based on the information collected from the kernel and return to the
     // caller the evaluated task's deadline.
     //
     // This method implements the main task ordering logic of the scheduler.
-    fn update_enqueued(&mut self, task: &QueuedTask) -> u64 {
-        // Get task information if the task is already stored in the task map,
-        // otherwise create a new entry for it.
-        let task_info = self.task_map.tasks.entry(task.pid).or_insert(TaskInfo {
-            vruntime: self.min_vruntime,
-        });
-
+    fn update_enqueued(&mut self, task: &mut QueuedTask) -> u64 {
         // Update global minimum vruntime based on the previous task's vruntime.
         if self.min_vruntime < task.vtime {
             self.min_vruntime = task.vtime;
         }
 
-        // Estimate the used time slice based on total runtime since the last sleep.
-        //
-        // Cap the value to slice_ns, since exec_runtime accumulates across multiple enqueue
-        // events, but what matters here is the time used in the most recent slice, so:
-        //  - if the task didn't sleep, it's the full slice_ns,
-        //  - if it did sleep, it's exec_runtime.
-        //
-        // Note that there may be some inaccuracies here, as a task can exceed its assigned time
-        // slice due to factors like holding locks or becoming non-deschedulable. These
-        // inaccuracies are tolerated to ensure smoother vruntime progression and prevent excessive
-        // gaps between tasks' vruntimes.
-        let slice = task.exec_runtime.min(self.slice_ns);
-
         // Update task's vruntime re-aligning it to min_vruntime (never allow a task to accumulate
         // a budget of more than a time slice to prevent starvation).
         let min_vruntime = self.min_vruntime.saturating_sub(self.slice_ns);
-        if task_info.vruntime < min_vruntime {
-            task_info.vruntime = min_vruntime;
+        if task.vtime == 0 {
+            // Slightly penalize new tasks by charging an extra time slice to prevent bursts of such
+            // tasks from disrupting the responsiveness of already running ones.
+            task.vtime = min_vruntime + Self::scale_by_task_weight_inverse(task, self.slice_ns);
+        } else if task.vtime < min_vruntime {
+            task.vtime = min_vruntime;
         }
-        let vslice = slice * 100 / task.weight;
-        task_info.vruntime += vslice;
+        task.vtime += Self::scale_by_task_weight_inverse(task, task.stop_ts - task.start_ts);
 
         // Return the task's deadline.
-        task_info.vruntime + task.exec_runtime.min(self.slice_ns * 100)
+        task.vtime + task.exec_runtime.min(self.slice_ns * 100)
+    }
+
+    // Dispatch the first task from the task pool (sending them to the BPF dispatcher).
+    //
+    // Return true on success, false if the BPF backend can't accept any more dispatch.
+    fn dispatch_task(&mut self) -> bool {
+        let nr_waiting = self.nr_tasks_waiting() + 1;
+
+        if let Some(task) = self.tasks.pop_first() {
+            // Scale time slice based on the amount of tasks that are waiting in the
+            // scheduler's queue and the previously unused time slice budget, but make sure
+            // to assign at least slice_us_min.
+            let slice_ns = (self.slice_ns / nr_waiting).max(self.slice_ns_min);
+
+            // Create a new task to dispatch.
+            let mut dispatched_task = DispatchedTask::new(&task.qtask);
+
+            // Assign the time slice to the task and propagate the vruntime.
+            dispatched_task.slice_ns = slice_ns;
+
+            // Propagate the evaluated task's deadline to the scx_rustland_core backend.
+            dispatched_task.vtime = task.deadline;
+
+            // Try to pick an idle CPU for the task.
+            let cpu = self
+                .bpf
+                .select_cpu(task.qtask.pid, task.qtask.cpu, task.qtask.flags);
+            dispatched_task.cpu = if cpu >= 0 {
+                // An idle CPU was found, dispatch the task there.
+                cpu
+            } else if self.opts.percpu_local && task.qtask.nr_cpus_allowed == 1 {
+                // Task is restricted to run on a single CPU, dispatch it to that one.
+                task.qtask.cpu
+            } else {
+                // No idle CPU found, dispatch to the first CPU available.
+                RL_CPU_ANY
+            };
+
+            // Send task to the BPF dispatcher.
+            if self.bpf.dispatch_task(&dispatched_task).is_err() {
+                // If dispatching fails, re-add the task to the pool and skip further dispatching.
+                self.tasks.insert(task);
+
+                return false;
+            }
+        }
+
+        return true;
     }
 
     // Drain all the tasks from the queued list, update their vruntime (Self::update_enqueued()),
@@ -325,13 +314,13 @@ impl<'a> Scheduler<'a> {
     fn drain_queued_tasks(&mut self) {
         loop {
             match self.bpf.dequeue_task() {
-                Ok(Some(task)) => {
+                Ok(Some(mut task)) => {
                     // Update task information and determine vruntime.
-                    let deadline = self.update_enqueued(&task);
+                    let deadline = self.update_enqueued(&mut task);
                     let timestamp = Self::now();
 
                     // Insert task in the task pool (ordered by vruntime).
-                    self.task_pool.push(Task {
+                    self.tasks.insert(Task {
                         qtask: task,
                         deadline,
                         timestamp,
@@ -346,88 +335,28 @@ impl<'a> Scheduler<'a> {
                 }
             }
         }
-    }
 
-    // Return the total amount of tasks that are waiting to be scheduled.
-    fn nr_tasks_waiting(&mut self) -> u64 {
-        let nr_queued = *self.bpf.nr_queued_mut();
-        let nr_scheduled = *self.bpf.nr_scheduled_mut();
-
-        nr_queued + nr_scheduled
-    }
-
-    // Dispatch the first task from the task pool (sending them to the BPF dispatcher).
-    fn dispatch_tasks(&mut self) {
-        match self.task_pool.pop() {
-            Some(task) => {
-                // Scale time slice based on the amount of tasks that are waiting in the
-                // scheduler's queue and the previously unused time slice budget, but make sure
-                // to assign at least slice_us_min.
-                let nr_waiting = self.nr_tasks_waiting() + 1;
-                let slice_ns = (self.slice_ns / nr_waiting).max(self.slice_ns_min);
-
-                // Create a new task to dispatch.
-                let mut dispatched_task = DispatchedTask::new(&task.qtask);
-
-                // Assign the time slice to the task and propagate the vruntime.
-                dispatched_task.slice_ns = slice_ns;
-
-                // Propagate the evaluated task's deadline to the scx_rustland_core backend.
-                dispatched_task.vtime = task.deadline;
-
-                // Try to pick an idle CPU for the task.
-                let cpu = self
-                    .bpf
-                    .select_cpu(task.qtask.pid, task.qtask.cpu, task.qtask.flags);
-                dispatched_task.cpu = if cpu >= 0 { cpu } else { RL_CPU_ANY };
-
-                // Send task to the BPF dispatcher.
-                match self.bpf.dispatch_task(&dispatched_task) {
-                    Ok(_) => {}
-                    Err(_) => {
-                        /*
-                         * Re-add the task to the dispatched list in case of failure and stop
-                         * dispatching.
-                         */
-                        self.task_pool.push(task);
-                    }
-                }
-            }
-            None => {}
-        }
+        // Dispatch the first task from the task pool.
+        self.dispatch_task();
     }
 
     // Main scheduling function (called in a loop to periodically drain tasks from the queued list
     // and dispatch them to the BPF part via the dispatched list).
     fn schedule(&mut self) {
         self.drain_queued_tasks();
-        self.dispatch_tasks();
 
         // Notify the dispatcher if there are still pending tasks to be processed,
-        self.bpf.notify_complete(self.task_pool.tasks.len() as u64);
+        self.bpf.notify_complete(self.tasks.len() as u64);
     }
 
-    // Get total page faults from /proc/self/stat.
+    // Get total page faults from the process.
     fn get_page_faults() -> Result<u64, io::Error> {
-        let path = format!("/proc/self/stat");
-        let mut file = File::open(path)?;
+        let myself = Process::myself().map_err(|e| io::Error::new(io::ErrorKind::Other, e))?;
+        let stat = myself
+            .stat()
+            .map_err(|e| io::Error::new(io::ErrorKind::Other, e))?;
 
-        // Read the contents of the file into a string.
-        let mut content = String::new();
-        file.read_to_string(&mut content)?;
-
-        // Parse the relevant fields and calculate the total page faults.
-        let fields: Vec<&str> = content.split_whitespace().collect();
-        if fields.len() >= 12 {
-            let minflt: u64 = fields[9].parse().unwrap_or(0);
-            let majflt: u64 = fields[11].parse().unwrap_or(0);
-            Ok(minflt + majflt)
-        } else {
-            Err(io::Error::new(
-                io::ErrorKind::InvalidData,
-                "Invalid format in /proc/[PID]/stat",
-            ))
-        }
+        Ok(stat.minflt + stat.majflt)
     }
 
     fn run(&mut self) -> Result<UserExitInfo> {
@@ -438,9 +367,8 @@ impl<'a> Scheduler<'a> {
             self.schedule();
 
             // Handle monitor requests asynchronously.
-            match req_ch.try_recv() {
-                Ok(()) => res_ch.send(self.get_metrics())?,
-                Err(_) => {}
+            if req_ch.try_recv().is_ok() {
+                res_ch.send(self.get_metrics())?;
             }
         }
 
@@ -462,7 +390,7 @@ fn main() -> Result<()> {
         println!(
             "{} version {} - scx_rustland_core {}",
             SCHEDULER_NAME,
-            VERSION,
+            build_id::full_version(env!("CARGO_PKG_VERSION")),
             scx_rustland_core::VERSION
         );
         return Ok(());
@@ -476,7 +404,9 @@ fn main() -> Result<()> {
     let loglevel = simplelog::LevelFilter::Info;
 
     let mut lcfg = simplelog::ConfigBuilder::new();
-    lcfg.set_time_level(simplelog::LevelFilter::Error)
+    lcfg.set_time_offset_to_local()
+        .expect("Failed to set local time offset")
+        .set_time_level(simplelog::LevelFilter::Error)
         .set_location_level(simplelog::LevelFilter::Off)
         .set_target_level(simplelog::LevelFilter::Off)
         .set_thread_level(simplelog::LevelFilter::Off);
