@@ -41,7 +41,7 @@
 
 #include <scx/common.bpf.h>
 
-#include <scx/bpf_arena_common.h>
+#include <scx/bpf_arena_common.bpf.h>
 #include <scx/bpf_arena_spin_lock.h>
 #include <scx/ravg_impl.bpf.h>
 
@@ -56,7 +56,6 @@
 #include "lb_domain.h"
 #include "deadline.h"
 
-#include <scx/bpf_arena_common.h>
 #include <errno.h>
 #include <stdbool.h>
 #include <bpf/bpf_core_read.h>
