@@ -503,11 +503,10 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
     }
 
     pub fn get_stats(&self) -> BTreeMap<usize, NodeStats> {
-        let mut stats = BTreeMap::new();
-        for node in self.nodes.iter() {
-            stats.insert(node.id, node.stats());
-        }
-        stats
+        self.nodes
+            .iter()
+            .map(|node| (node.id, node.stats()))
+            .collect()
     }
 
     fn create_domain_hierarchy(&mut self) -> Result<()> {
