@@ -29,6 +29,8 @@ use std::time::UNIX_EPOCH;
 use stats::ClusterStats;
 use stats::NodeStats;
 
+use std::ffi::c_ulong;
+
 #[macro_use]
 extern crate static_assertions;
 
@@ -53,7 +55,9 @@ use scx_utils::scx_ops_load;
 use scx_utils::scx_ops_open;
 use scx_utils::uei_exited;
 use scx_utils::uei_report;
+use scx_utils::Core;
 use scx_utils::Cpumask;
+use scx_utils::Llc;
 use scx_utils::Topology;
 use scx_utils::UserExitInfo;
 use scx_utils::NR_CPU_IDS;
@@ -124,14 +128,6 @@ struct Opts {
     #[clap(short = 'c', long, default_value = "3")]
     cache_level: u32,
 
-    /// Instead of using cache locality, set the cpumask for each domain
-    /// manually. Provide multiple --cpumasks, one for each domain. E.g.
-    /// --cpumasks 0xff_00ff --cpumasks 0xff00 will create two domains, with
-    /// the corresponding CPUs belonging to each domain. Each CPU must
-    /// belong to precisely one domain.
-    #[clap(short = 'C', long, num_args = 1.., conflicts_with = "cache_level")]
-    cpumasks: Vec<String>,
-
     /// When non-zero, enable greedy task stealing. When a domain is idle, a cpu
     /// will attempt to steal tasks from another domain as follows:
     ///
@@ -365,6 +361,169 @@ struct Scheduler<'a> {
 }
 
 impl<'a> Scheduler<'a> {
+    fn setup_allocators(skel: &mut BpfSkel<'a>) -> Result<()> {
+        // Allocate the arena memory from the BPF side so userspace initializes it before starting
+        // the scheduler. Despite the function call's name this is neither a test nor a test run,
+        // it's the recommended way of executing SEC("syscall") probes.
+        let mut args = types::arena_init_args {
+            static_pages: bpf_intf::consts_STATIC_ALLOC_PAGES_GRANULARITY as c_ulong,
+            task_ctx_size: std::mem::size_of::<types::task_ctx>() as c_ulong,
+        };
+
+        let input = ProgramInput {
+            context_in: Some(unsafe {
+                std::slice::from_raw_parts_mut(
+                    &mut args as *mut _ as *mut u8,
+                    std::mem::size_of_val(&args),
+                )
+            }),
+            ..Default::default()
+        };
+
+        let output = skel.progs.arena_init.test_run(input)?;
+        if output.return_value != 0 {
+            bail!(
+                "Could not initialize arenas, p2dq_setup returned {}",
+                output.return_value as i32
+            );
+        }
+
+        Ok(())
+    }
+
+    fn setup_topology_node(
+        skel: &mut BpfSkel<'a>,
+        mask: &[u64],
+        data_size: usize,
+        id: usize,
+    ) -> Result<()> {
+        let mut args = types::arena_alloc_mask_args {
+            bitmap: 0 as c_ulong,
+        };
+
+        let input = ProgramInput {
+            context_in: Some(unsafe {
+                std::slice::from_raw_parts_mut(
+                    &mut args as *mut _ as *mut u8,
+                    std::mem::size_of_val(&args),
+                )
+            }),
+            ..Default::default()
+        };
+
+        let output = skel.progs.arena_alloc_mask.test_run(input)?;
+        if output.return_value != 0 {
+            bail!(
+                "Could not initialize arenas, setup_topology_node returned {}",
+                output.return_value as i32
+            );
+        }
+
+        let ptr = unsafe { std::mem::transmute::<u64, &mut [u64; 10]>(args.bitmap) };
+
+        let (valid_mask, _) = ptr.split_at_mut(mask.len());
+        valid_mask.clone_from_slice(mask);
+
+        let mut args = types::arena_topology_node_init_args {
+            bitmap: args.bitmap as c_ulong,
+            data_size: data_size as c_ulong,
+            id: id as c_ulong,
+        };
+
+        let input = ProgramInput {
+            context_in: Some(unsafe {
+                std::slice::from_raw_parts_mut(
+                    &mut args as *mut _ as *mut u8,
+                    std::mem::size_of_val(&args),
+                )
+            }),
+            ..Default::default()
+        };
+        let output = skel.progs.arena_topology_node_init.test_run(input)?;
+        if output.return_value != 0 {
+            bail!(
+                "p2dq_topology_node_init returned {}",
+                output.return_value as i32
+            );
+        }
+
+        Ok(())
+    }
+
+    fn setup_topology(skel: &mut BpfSkel<'a>) -> Result<()> {
+        let topo = Topology::new().expect("Failed to build host topology");
+
+        // We never use the topology-provided IDs, because we do not need them anymore now that
+        // we have a proper topology struct. We instead only use sequential IDs we create
+        // ourselves to attach to the topology nodes. This is because the Topology-provided IDs
+        // are used to determine relations between topology nodes, e.g., LLCs belonging to a
+        // node, while sequential IDs are useful for linearly scanning the topology, e.g.,
+        // iterating over domains. This is why we LLC IDs and domain IDs in scx_wd40 are different.
+        // For now we only need the sequential IDs, so use those. Eventually we will be able
+        // to remove those, too, once we remove the hardcoded arrays from the code.
+        Self::setup_topology_node(skel, topo.span.as_raw_slice(), 0, 0)?;
+
+        for (id, (_, node)) in topo.nodes.into_iter().enumerate() {
+            Self::setup_topology_node(skel, node.span.as_raw_slice(), 0, id)?;
+        }
+
+        for (id, (_, llc)) in topo.all_llcs.into_iter().into_iter().enumerate() {
+            Self::setup_topology_node(
+                skel,
+                Arc::<Llc>::into_inner(llc)
+                    .expect("missing llc")
+                    .span
+                    .as_raw_slice(),
+                0,
+                id,
+            )?;
+        }
+        for (id, (_, core)) in topo.all_cores.into_iter().into_iter().enumerate() {
+            Self::setup_topology_node(
+                skel,
+                Arc::<Core>::into_inner(core)
+                    .expect("missing core")
+                    .span
+                    .as_raw_slice(),
+                0,
+                id,
+            )?;
+        }
+        for (id, (_, cpu)) in topo.all_cpus.into_iter().into_iter().enumerate() {
+            let mut mask = [0; 9];
+            mask[cpu.id.checked_shr(64).unwrap_or(0)] |= 1 << (cpu.id % 64);
+            Self::setup_topology_node(skel, &mask, 0, id)?;
+        }
+
+        Ok(())
+    }
+
+    fn setup_wd40(skel: &mut BpfSkel<'a>) -> Result<()> {
+        // Allocate the arena memory from the BPF side so userspace initializes it before starting
+        // the scheduler. Despite the function call's name this is neither a test nor a test run,
+        // it's the recommended way of executing SEC("syscall") probes.
+        let input = ProgramInput {
+            ..Default::default()
+        };
+        let output = skel.progs.wd40_setup.test_run(input)?;
+        if output.return_value != 0 {
+            bail!(
+                "Could not initialize WD40 arenas, wd40_arena_setup returned {}",
+                output.return_value as i32
+            );
+        }
+
+        Ok(())
+    }
+
+    fn setup_arenas(skel: &mut BpfSkel<'a>) -> Result<()> {
+        Self::setup_allocators(skel)?;
+        Self::setup_topology(skel)?;
+        Self::setup_wd40(skel)?;
+
+        Ok(())
+    }
+
     fn init(opts: &Opts, open_object: &'a mut MaybeUninit<OpenObject>) -> Result<Self> {
         // Open the BPF prog first for verification.
         let mut skel_builder = BpfSkelBuilder::default();
@@ -377,7 +536,7 @@ impl<'a> Scheduler<'a> {
         let mut skel = scx_ops_open!(skel_builder, open_object, wd40).unwrap();
 
         // Initialize skel according to @opts.
-        let domains = Arc::new(DomainGroup::new(&Topology::new()?, &opts.cpumasks)?);
+        let domains = Arc::new(DomainGroup::new(&Topology::new()?)?);
 
         if *NR_CPU_IDS > MAX_CPUS {
             bail!(
@@ -401,22 +560,6 @@ impl<'a> Scheduler<'a> {
         skel.maps.rodata_data.nr_doms = domains.nr_doms() as u32;
         skel.maps.rodata_data.nr_cpu_ids = *NR_CPU_IDS as u32;
 
-        // Any CPU with dom > MAX_DOMS is considered offline by default. There
-        // are a few places in the BPF code where we skip over offlined CPUs
-        // (e.g. when initializing or refreshing tune params), and elsewhere the
-        // scheduler will error if we try to schedule from them.
-        for cpu in 0..*NR_CPU_IDS {
-            skel.maps.rodata_data.cpu_dom_id_map[cpu] = u32::MAX;
-        }
-
-        for (id, dom) in domains.doms().iter() {
-            for cpu in dom.mask().iter() {
-                skel.maps.rodata_data.cpu_dom_id_map[cpu] = (*id)
-                    .try_into()
-                    .expect("Domain ID could not fit into 32 bits");
-            }
-        }
-
         if opts.partial {
             skel.struct_ops.wd40_mut().flags |= *compat::SCX_OPS_SWITCH_PARTIAL;
         }
@@ -434,22 +577,11 @@ impl<'a> Scheduler<'a> {
 
         let mut skel = scx_ops_load!(skel, wd40, uei)?;
 
-        // Allocate the arena memory from the BPF side so userspace initializes it before starting
-        // the scheduler. Despite the function call's name this is neither a test nor a test run,
-        // it's the recommended way of executing SEC("syscall") probes.
-        let input = ProgramInput {
-            ..Default::default()
-        };
-        let output = skel.progs.wd40_arena_setup.test_run(input)?;
-        if output.return_value != 0 {
-            bail!(
-                "Could not initialize WD40 arenas, wd40_arena_setup returned {}",
-                output.return_value as i32
-            );
-        }
+        Self::setup_arenas(&mut skel)?;
+        
 
-        println!(
-            "Mask length {} NR_CPU_IDS {}",
+        info!(
+            "Mask length {}, number of possible CPUs {}",
             skel.maps.bss_data.mask_size, skel.maps.rodata_data.nr_cpu_ids
         );
         // Read the mask length chosen by BPF. We count elements in the u64 array, like the BPF
@@ -458,6 +590,8 @@ impl<'a> Scheduler<'a> {
         // This invocation is safe because there is no concurrency in the program during initialization.
         unsafe { MASK_LEN = skel.maps.bss_data.mask_size as usize };
 
+        let types::topo_level(index) = types::topo_level::TOPO_LLC;
+
         for numa in 0..domains.nr_nodes() {
             let mut numa_mask = Cpumask::new();
             let node_domains = domains.numa_doms(&numa);
@@ -470,7 +604,9 @@ impl<'a> Scheduler<'a> {
             info!("NODE[{:02}] mask= {}", numa, numa_mask);
 
             for dom in node_domains.iter() {
-                let domc = unsafe { &mut *skel.maps.bss_data.dom_ctxs[dom.id()] };
+                // XXX Remove this by using the topo node's cpumask.
+                let ptr = skel.maps.bss_data.topo_nodes[index as usize][dom.id()];
+                let domc = unsafe { std::mem::transmute::<u64, &mut types::dom_ctx>(ptr) };
                 update_bpf_mask(domc.cpumask, &dom.mask())?;
 
                 skel.maps.bss_data.dom_numa_id_map[dom.id()] =
@@ -487,7 +623,9 @@ impl<'a> Scheduler<'a> {
         for (id, dom) in domains.doms().iter() {
             let mut ctx = dom.ctx.lock().unwrap();
 
-            *ctx = Some(skel.maps.bss_data.dom_ctxs[*id]);
+            let ptr = skel.maps.bss_data.topo_nodes[index as usize][*id];
+            let domc = unsafe { std::mem::transmute::<u64, &mut types::dom_ctx>(ptr) };
+            *ctx = Some(domc);
         }
 
         info!("WD40 scheduler started! Run `scx_wd40 --monitor` for metrics.");
@@ -647,7 +785,7 @@ impl<'a> Scheduler<'a> {
             }
         }
 
-        self.struct_ops.take();
+        let _ = self.struct_ops.take();
         uei_report!(&self.skel, uei)
     }
 }
@@ -682,7 +820,9 @@ fn main() -> Result<()> {
         _ => simplelog::LevelFilter::Trace,
     };
     let mut lcfg = simplelog::ConfigBuilder::new();
-    lcfg.set_time_level(simplelog::LevelFilter::Error)
+    lcfg.set_time_offset_to_local()
+        .expect("Failed to set local time offset")
+        .set_time_level(simplelog::LevelFilter::Error)
         .set_location_level(simplelog::LevelFilter::Off)
         .set_target_level(simplelog::LevelFilter::Off)
         .set_thread_level(simplelog::LevelFilter::Off);
