@@ -62,7 +62,7 @@ pub struct TuiArgs {
     #[arg(long, default_value_t = -1)]
     pub process_id: i32,
     /// Custom perf events colon delimited (ex: "<event_name>:<event and umask ex: 0x023>:<event_type ex: 4>")
-    #[arg(long)]
+    #[arg(long, num_args = 1.., value_parser)]
     pub perf_events: Vec<String>,
     /// Default perf event colon delimited (ex: "<event_name>:<event and umask ex: 0x023>:<event_type ex: 4>")
     #[arg(long, default_value = "hw:cycles")]
@@ -113,6 +113,9 @@ pub struct TraceArgs {
     /// times to increase verbosity.
     #[clap(short = 'v', long, action = clap::ArgAction::Count)]
     pub verbose: u8,
+    /// Add a list of kprobe events to the trace.
+    #[clap(short = 'k', long, num_args = 1.., value_parser)]
+    pub kprobes: Vec<String>,
 }
 
 #[allow(clippy::large_enum_variant)]
