@@ -38,7 +38,8 @@ impl Default for KeyMap {
     fn default() -> Self {
         let mut bindings = HashMap::new();
         bindings.insert(Key::Char('d'), Action::SetState(AppState::Default));
-        bindings.insert(Key::Char('e'), Action::SetState(AppState::Event));
+        bindings.insert(Key::Char('e'), Action::SetState(AppState::PerfEvent));
+        bindings.insert(Key::Char('K'), Action::SetState(AppState::KprobeEvent));
         bindings.insert(Key::Char('f'), Action::ToggleCpuFreq);
         bindings.insert(Key::Char('u'), Action::ToggleUncoreFreq);
         bindings.insert(Key::Char('L'), Action::ToggleLocalization);
@@ -327,12 +328,14 @@ pub fn parse_key(key_str: &str) -> Result<Key> {
     {
         Ok(Key::Code(keycode_wrapper.into()))
     } else {
-        match key_str {
-            "Page Up" | "PageUp" => Ok(Key::Code(KeyCode::PageUp)),
-            "Page Down" | "PageDown" => Ok(Key::Code(KeyCode::PageDown)),
-            "Up" => Ok(Key::Code(KeyCode::Up)),
-            "Down" => Ok(Key::Code(KeyCode::Down)),
-            "Enter" => Ok(Key::Code(KeyCode::Enter)),
+        match key_str.to_lowercase().as_str() {
+            "page up" | "pageup" => Ok(Key::Code(KeyCode::PageUp)),
+            "page down" | "pagedown" => Ok(Key::Code(KeyCode::PageDown)),
+            "up" => Ok(Key::Code(KeyCode::Up)),
+            "down" => Ok(Key::Code(KeyCode::Down)),
+            "enter" => Ok(Key::Code(KeyCode::Enter)),
+            "backspace" => Ok(Key::Code(KeyCode::Backspace)),
+            "esc" | "escape" => Ok(Key::Code(KeyCode::Esc)),
             _ => Err(anyhow!("Invalid key: {}", key_str)),
         }
     }
@@ -342,7 +345,8 @@ pub fn parse_key(key_str: &str) -> Result<Key> {
 pub fn parse_action(action_str: &str) -> Result<Action> {
     match action_str {
         "AppStateDefault" => Ok(Action::SetState(AppState::Default)),
-        "AppStateEvent" => Ok(Action::SetState(AppState::Event)),
+        "AppStatePerfEvent" => Ok(Action::SetState(AppState::PerfEvent)),
+        "AppStateKprobeEvent" => Ok(Action::SetState(AppState::KprobeEvent)),
         "ToggleCpuFreq" => Ok(Action::ToggleCpuFreq),
         "ToggleUncoreFreq" => Ok(Action::ToggleUncoreFreq),
         "ToggleLocalization" => Ok(Action::ToggleLocalization),
@@ -369,6 +373,8 @@ pub fn parse_action(action_str: &str) -> Result<Action> {
         "PageDown" => Ok(Action::PageDown),
         "PageUp" => Ok(Action::PageUp),
         "Enter" => Ok(Action::Enter),
+        "Esc" => Ok(Action::Esc),
+        "Backspace" => Ok(Action::Backspace),
         _ => Err(anyhow!("Invalid action: {}", action_str)),
     }
 }
